{% extends "layout.html" %}
{% block title %}{{ _('Warning for %(username)s lifted', username=username) }}{% endblock %}
{% block preheader %}{{ _('Your appeal has been rejected') }}.{% endblock %}
{% block content %}<p>{% if action_type == "user_suspension" %}{{ _('Your appeal on your account suspension has been rejected') }}{% elif action_type == "user_warning" %}{{ _('Your appeal on your warning has been rejected') }}{% else %}{{ _('Your appeal on the following content suspension has been rejected') }}{% endif %}.</p>
                        {% if action_type != "user_suspension" and comment_url %}
                        <p>{{ _('Comment:') }}</p>
                        <table class="body-content" width="100%" cellpadding="0" cellspacing="0" role="presentation">
                          <tr>
                            <td>
                              <img class="user-picture" src="{{ user_image_url }}" alt="">
                              <strong>{{ username }}</strong>
                            </td>
                           </tr>
                          <tr>
                            <td>{{ text }}</td>
                          </tr>
                          <tr>
                            <td class="content-date"><a href="{{ comment_url }}" target="_blank">{{ created_at }}</a></td>
                          </tr>
                        </table>{% endif %}
                        {% if action_type != "user_suspension" and workout_url %}
                        <p>{{ _('Workout:') }}</p>
                        <table class="body-content" width="100%" cellpadding="0" cellspacing="0" role="presentation">
                          <tr>
                            <td>
                              <img class="user-picture" src="{{ user_image_url }}" alt="">
                              <strong>{{ username }}</strong>
                            </td>
                          </tr>
                          <tr>
                            <td>{{ title }}</td>
                          </tr>
                          {% if map %}<tr>
                            <td><img class="map" src="{{ map }}" alt=""></td>
                           </tr>{% endif %}
                          <tr>
                            <td class="content-date"><a href="{{ workout_url }}" target="_blank">{{ workout_date }}</a></td>
                          </tr>
                        </table>{% endif %}{% endblock %}